#include <stdio.h>
#include <signal.h>
#include <execinfo.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>

//!!!ONLY STATIC TYPES FOR MEM LEAK DIAGNOSTIC CODE

#define STACK_TRACE_SIZE 1024
#define MEMLEAKS_ARRAY_SIZE 4096

typedef struct 
{
  unsigned int addr;
  int size;
  char stackTrace[STACK_TRACE_SIZE];
} MemLeak;

//global variable for memory leaks
MemLeak memLeaks[MEMLEAKS_ARRAY_SIZE];
int currArrSize = 0;

//holes in memLeaks
int arrHoles[MEMLEAKS_ARRAY_SIZE];
int arrHolesCurrPos = -1;

//works similar as push on stack
int addHole(int holeIndex)
{
  if (arrHolesCurrPos < MEMLEAKS_ARRAY_SIZE)
  {
    arrHoles[++arrHolesCurrPos] = holeIndex;
    return 1;
  }
  return 0;
}

//works similar as pop on stack
int removeHole(int* holeIndex)
{
  if (arrHolesCurrPos >= 0)
  {
    *holeIndex = arrHoles[arrHolesCurrPos--];
    return 1;
  }	
  return 0;
}

//finds address position in mem leaks array
int findAddrInMemLekasArr(int addr)
{
  int i;
  
  for(i=0; i<MEMLEAKS_ARRAY_SIZE; ++i)
  {
    if(memLeaks[i].addr == addr)
	  return i;
  }
  
  return -1;
}

//gets call stack
void getCallStack(char* retVal)
{
  void *trace[16];
  char **messages = (char **)NULL;
  int i, trace_size, msg_size;
  trace_size = backtrace(trace, 16);
  messages = backtrace_symbols(trace, trace_size);
  int pos = 0;
  
  for (i=0; i<trace_size; ++i)
  {
    msg_size = strlen(messages[i]);
    
    //string too long
    if(pos + msg_size >= STACK_TRACE_SIZE) 
      break;
    
    strncpy(retVal + pos, messages[i], msg_size);
    retVal[pos + msg_size] = '\n';
    pos += (msg_size + 1);
  }
   
  retVal[pos] = 0;
  free(messages);
}

//adds element to array
void addToMemLeaksArr(int addr, int size, char *tr)
{
   int index;
    
   if (!removeHole(&index))
     index = currArrSize;
	 	
   if((0 <= index) && (MEMLEAKS_ARRAY_SIZE > index))
   {   
       memLeaks[index].addr = addr;
       memLeaks[index].size = size;
       strncpy(memLeaks[index].stackTrace, tr, STACK_TRACE_SIZE);
       ++currArrSize;
   }
   else
       fprintf(stderr, "MEMLEAKS_ARRAY_SIZE is too small\n");
}

//removes address from array
void removeFromMemLeaksArr(int addr)
{
  int index = findAddrInMemLekasArr(addr);
  if((0 <= index) && (MEMLEAKS_ARRAY_SIZE > index))
  {
    memLeaks[index].addr = 0;
    --currArrSize;
    addHole(index);
  }
}

//your overloaded malloc or new operator
void* diagnosticMalloc(size_t size)
{
   void *allocMem = malloc(size);
   
   //puts stack trace into array
   char stackTrace[STACK_TRACE_SIZE];
   getCallStack(stackTrace);
   addToMemLeaksArr((int)allocMem, size, stackTrace);
   
   return allocMem; 
}

//your overloaded free or delete operator
void diagnosticFree(void *ptr)
{
   removeFromMemLeaksArr((int)ptr);
   
   free(ptr);
}

//prints leaks to stderr
void printMemoryLeaks()
{
  int i;
  for(i=0; i<MEMLEAKS_ARRAY_SIZE; ++i)
  {
     if(memLeaks[i].addr > 0)
     {
         fprintf(stderr, "Memory leak found:\naddr: %p\nsize: %d\ncall stack:\n%s\n",
                 memLeaks[i].addr, memLeaks[i].size, memLeaks[i].stackTrace);
     }
  } 
}

//signal handler
void signalCallbackHandler(int signum)
{
  printMemoryLeaks();
}

void generateMemLeak()
{
  //world is cool :) ... 
  double *ptr = (double*)diagnosticMalloc(sizeof(double));  
}

int main()
{  
   //register callback for:  kill -2 pid
   signal(SIGINT, signalCallbackHandler);
   
   generateMemLeak();
   
   //malloc and free
   int *ptr = (int*)diagnosticMalloc(sizeof(int));
   *ptr = 5;
   diagnosticFree(ptr);
   
   while(1); //endless loop      
}